/* THIS IS A GENERATED FILE. Edit generate_codebook.c and its input */

/*
 * This intermediary file and the files that used to create it are under 
 * The LGPL. See the file COPYING.
 */

#include "defines.h"

  /* /home/david/codec2-dev/src/codebook/newamp1_energy_q.txt */
static const float codes0[] = {
  10,
  12.5,
  15,
  17.5,
  20,
  22.5,
  25,
  27.5,
  30,
  32.5,
  35,
  37.5,
  40,
  42.5,
  45,
  47.5
};

const struct lsp_codebook newamp1_energy_cb[] = {
  /* /home/david/codec2-dev/src/codebook/newamp1_energy_q.txt */
  {
    1,
    4,
    16,
    codes0
  },
  { 0, 0, 0, 0 }
};
