/* THIS IS A GENERATED FILE. Edit generate_codebook.c and its input */

/*
 * This intermediary file and the files that used to create it are under
 * The LGPL. See the file COPYING.
 */

#include "defines.h"

  /* codebook/lspdt1.txt */
static const float codes0[] = {
  -75,
  -50,
  -25,
  0,
  25,
  50,
  75,
  100
};
  /* codebook/lspdt2.txt */
static const float codes1[] = {
  -75,
  -50,
  -25,
  0,
  25,
  50,
  75,
  100
};
  /* codebook/lspdt3.txt */
static const float codes2[] = {
  -50,
  0,
  50,
  100
};
  /* codebook/lspdt4.txt */
static const float codes3[] = {
  -50,
  0,
  50,
  100
};
  /* codebook/lspdt5.txt */
static const float codes4[] = {
  -50,
  0,
  50,
  100
};
  /* codebook/lspdt6.txt */
static const float codes5[] = {
  -50,
  0,
  50,
  100
};
  /* codebook/lspdt7.txt */
static const float codes6[] = {
  -50,
  50
};
  /* codebook/lspdt8.txt */
static const float codes7[] = {
  -50,
  50
};
  /* codebook/lspdt9.txt */
static const float codes8[] = {
  -50,
  50
};
  /* codebook/lspdt10.txt */
static const float codes9[] = {
  -50,
  50
};

const struct lsp_codebook lsp_cbdt[] = {
  /* codebook/lspdt1.txt */
  {
    1,
    3,
    8,
    codes0
  },
  /* codebook/lspdt2.txt */
  {
    1,
    3,
    8,
    codes1
  },
  /* codebook/lspdt3.txt */
  {
    1,
    2,
    4,
    codes2
  },
  /* codebook/lspdt4.txt */
  {
    1,
    2,
    4,
    codes3
  },
  /* codebook/lspdt5.txt */
  {
    1,
    2,
    4,
    codes4
  },
  /* codebook/lspdt6.txt */
  {
    1,
    2,
    4,
    codes5
  },
  /* codebook/lspdt7.txt */
  {
    1,
    1,
    2,
    codes6
  },
  /* codebook/lspdt8.txt */
  {
    1,
    1,
    2,
    codes7
  },
  /* codebook/lspdt9.txt */
  {
    1,
    1,
    2,
    codes8
  },
  /* codebook/lspdt10.txt */
  {
    1,
    1,
    2,
    codes9
  },
  { 0, 0, 0, 0 }
};
