/* Generated by write_pilot_file() Octave function */

#define NSYMROW      4   /* number of data symbols per carrier (number of rows)     */
#define NS           4   /* number of data symbols between pilots                   */
#define NPILOTSFRAME 2   /* number of pilot symbols per carrier                     */
#define PILOTS_NC    7   /* number of carriers                                      */

#define NSYMROWPILOT 6   /* number of rows after pilots inserted                    */

