/* THIS IS A GENERATED FILE. Edit generate_codebook.c and its input */

/*
 * This intermediary file and the files that used to create it are under
 * The LGPL. See the file COPYING.
 */

#include "defines.h"

  /* ../src/codebook/lspres_centre1.txt */
static const float codes0[] = {
  300,
  350,
  400,
  450,
  500,
  550,
  600,
  650
};
  /* ../src/codebook/lspres_bw1.txt */
static const float codes1[] = {
  35,
  80,
  140,
  250
};
  /* ../src/codebook/lsp3.txt */
static const float codes2[] = {
  500,
  550,
  600,
  650,
  700,
  750,
  800,
  850,
  900,
  950,
  1000,
  1050,
  1100,
  1150,
  1200,
  1250
};
  /* ../src/codebook/lsp4.txt */
static const float codes3[] = {
  700,
  800,
  900,
  1000,
  1100,
  1200,
  1300,
  1400,
  1500,
  1600,
  1700,
  1800,
  1900,
  2000,
  2100,
  2200
};

const struct lsp_codebook lsp_cbres[] = {
  /* ../src/codebook/lspres_centre1.txt */
  {
    1,
    3,
    8,
    codes0
  },
  /* ../src/codebook/lspres_bw1.txt */
  {
    1,
    2,
    4,
    codes1
  },
  /* ../src/codebook/lsp3.txt */
  {
    1,
    4,
    16,
    codes2
  },
  /* ../src/codebook/lsp4.txt */
  {
    1,
    4,
    16,
    codes3
  },
  { 0, 0, 0, 0 }
};
